import struct

import auxlib
import efalib
from auxlib import Enum, AuxSession

# For Python 2.7 compatibility, map input() to raw_input()
try:
    input = raw_input
except NameError:
    pass # We are probably running Python 3, so don't do anything



ADDR_DELTA_T = 0x32

HeaterState = Enum(
        OFF = 0,
        ON = 1,
        USERON = 2 # Not implemented
        )

HeaterMode = Enum(
        MANUAL = 1,  # Only supported mode as of this writing
        RELATIVE = 2,
        ABSOLUTE = 3,
        OVERRIDE = 4
        )

# Error codes reported in the heater status response
HeaterError = Enum(
        NONE = 0x80,           # No error
        USERMODEACTIVE = 0x81, # User Mode is active
        INVALIDHEATER = 0x82,  # Invalid heater number
        SETPOINTRANGE = 0x83,  # Setpoint out of range
        PWMPERIOD = 0x84,      # PWM period invalid
        PWMDUTYCYCLE = 0x85    # Duty cycle invalid
        )

# Commands recognized by the Delta-T.
# Note that there is some overlap with the EFA 
# (for example, TEMP_GET, GET_VERSION)
Command = Enum(
        TEMP_GET = 0x26,
        GET_TEMP_SENSOR_ROM = 0x29,
        GET_TEMP_SENSOR_SCRATCHPAD = 0x2A,
        GET_NUM_TEMP_SENSORS       = 0x2B,
        COH_NUMHEATERS  = 0xB0, 
        COH_ON_MANUAL   = 0xB1,
        COH_OFF         = 0xB4,
        COH_REPORT      = 0xB5,
        COH_TSASSOCIATE = 0xB6,
        COH_RESCAN      = 0xBF,
        GET_VERSION = 0xFE,
        )

class DeltaT:
    def __init__(self, comPort):
        self._session = AuxSession()
        self._session.openSerial(comPort, useRtsCts=False)

    def close(self):
        self._session.close()

    def getHeaterCount(self):
        packet = self._session.sendReceive(ADDR_DELTA_T,
                                           Command.COH_NUMHEATERS)
        return packet.parseData()


    def heaterOn(self, heaterIndex, periodSeconds, dutyCyclePercent):
        """
        heaterIndex: 0 thru NumHeaters-1
        """

        periodHundredthsSecond = int(periodSeconds * 100)

        byte0 = heaterIndex
        byte1 = periodHundredthsSecond & 0xFF
        byte2 = (periodHundredthsSecond >> 8) & 0xFF
        byte3 = dutyCyclePercent

        packet = self._session.sendReceive(ADDR_DELTA_T,
                                           Command.COH_ON_MANUAL,
                                           byte0,
                                           byte1,
                                           byte2,
                                           byte3)

        errorCode = packet.parseData()
        return errorCode
        
    def heaterOff(self, heaterIndex):
        """
        heaterIndex: 0 thru NumHeaters-1
        """

        packet = self._session.sendReceive(ADDR_DELTA_T,
                                           Command.COH_OFF,
                                           heaterIndex)

        errorCode = packet.parseData()
        return errorCode

    def getReport(self, heaterIndex):
        """
        heaterIndex: 0 thru NumHeaters-1
        """

        packet = self._session.sendReceive(ADDR_DELTA_T,
                                           Command.COH_REPORT,
                                           heaterIndex)

        fields = struct.unpack("<BBBHBhhhb", packet.dataAsByteString())

        report = DeltaTReport()

        report.errorCode = fields[0]
        report.state = fields[1]
        report.mode = fields[2]
        report.setpoint = fields[3]
        report.associatedTempSensorId = fields[4]
        report.associatedTempRaw = fields[5]
        report.ambientTempRaw = fields[6]
        report.periodRaw = fields[7]
        report.dutyCyclePercent = fields[8]

        return report

    def getVersion(self):
        packet = self._session.sendReceive(ADDR_DELTA_T,
                                           Command.GET_VERSION)

        major = packet.data[0]
        minor = packet.data[1]
        build = packet.data[2]*256 + packet.data[3]  # = YYDDD where YY = year, DDD = day of year

        return (major, minor, build)

    def getTempSensorRom(self, sensorIndex):
        packet = self._session.sendReceive(ADDR_DELTA_T,
                                           Command.GET_TEMP_SENSOR_ROM,
                                           sensorIndex)
        # Byte 0: Family code
        # Bytes 1-6: Unique ID
        # Byte 7: CRC code

        return packet

    def getTempSensorScratchpad(self, sensorIndex):
        packet = self._session.sendReceive(ADDR_DELTA_T,
                                           Command.GET_TEMP_SENSOR_SCRATCHPAD,
                                           sensorIndex)
        # Byte 0: Temperature LSB
        # Byte 1: Temperature MSB
        # Byte 2: High Temperature Alarm register or User Byte 1 (used by EFA for addressing)
        # Byte 3: Low Temperature Alarm register or User Byte 2
        # Byte 4: Configuration register
        #   0x7F = 12-bit resolution (1/16 degree increment; startup default)
        #   0x5F = 11-bit resolution (1/8 degree increment)
        #   0x3F = 10-bit resolution (1/4 degree increment)
        #   0x1F =  9-bit resolution (1/2 degree increment)
        # Byte 5: Reserved (0xFF)
        # Byte 6: Reserved (0xFF; have also seen 0x0F)
        # Byte 7: Reserved (0x10)
        # Byte 8: CRC
        # See DS18B20 datasheet for more details

        return packet

    def getNumTempSensors(self):
        packet = self._session.sendReceive(ADDR_DELTA_T,
                                           Command.GET_NUM_TEMP_SENSORS)
        return packet.parseData()

    def rescanForTempSensors(self):
        """
        Use this after temp sensors are plugged/unplugged to
        enumerate the current set of temp sensors

        Returns number of temp sensors found after re-scan
        """

        packet = self._session.sendReceive(ADDR_DELTA_T,
                                           Command.COH_RESCAN)
        return packet.parseData()

    def getTemperature(self, tempAddress):
        """
        tempAddress:
          0 = Primary
          1 = Ambient
          2 = Secondary
          3 = Backplate
          4 = M3
        """

        packet = self._session.sendReceive(ADDR_DELTA_T,
                                             Command.TEMP_GET,
                                             tempAddress)

        if tuple(packet.data) == (127, 127):
            return None # no temp sensor at this address
        else:
            # Temperature for Delta-T is returned MSB-first
            # (in contrast to EFA, which is LSB-first I think)
            return (256*packet.data[0] + packet.data[1]) / 16.0


class DeltaTReport:
    """
    Represent the response of the COH_REPORT command
    """

    def __init__(self):
        self.errorCode = None
        self.state = None
        self.mode = None
        self.setpoint = None
        self.associatedTempSensorId = None
        self.associatedTempRaw = None
        self.ambientTempRaw = None
        self.periodRaw = None
        self.dutyCyclePercent = None

    def associatedTempC(self):
        return efalib.rawTemperatureToCelsius(self.associatedTempRaw)

    def ambientTempC(self):
        return efalib.rawTemperatureToCelsius(self.ambientTempRaw)

    def periodSeconds(self):
        return self.periodRaw / 100.0

    def __str__(self):
        fields = [
            "Error code: %s" % HeaterError.getName(self.errorCode),
            "Heater state: %s" % HeaterState.getName(self.state),
            "Heater mode: %s" % HeaterMode.getName(self.mode),

            # Provisioned in protocol but not supported:
            #"Setpoint: %d" % self.setpoint,
            #"Associated sensor: %s" % self.associatedTempSensorId,
            #"Associated temp: %.2f C (%d)" % (self.associatedTempC(), self.associatedTempRaw),
            #"Ambient temp: %.2f C (%d)" % (self.ambientTempC(), self.ambientTempRaw),

            "Period: %.2f sec (%d)" % (self.periodSeconds(), self.periodRaw),
            "Duty cycle: %d%%" % (self.dutyCyclePercent),
        ]

        return "\n".join(fields)


def main():
    """
    Sample usage of the Delta-T library
    """

    deltat = DeltaT("COM4")
    deltat._session.abortOnTimeout = True

    version = deltat.getVersion()
    print("Version:", version)

    #heaterCount = deltat.getHeaterCount()
    #print "Number of heaters:", heaterCount

    showReports(deltat)

    print("Press Enter to turn on the primary heater at 30% power, cycling each second")
    print("(on for 0.3 sec, off for 0.7 sec)")
    input()
    deltat.heaterOn(0, 1.0, 30)

    showReports(deltat)

    print("Press Enter to turn on the secondary heater at 10% power, cycling each second")
    print("(on for 0.1 sec, off for 0.9 sec)")
    input()
    deltat.heaterOn(1, 1.0, 10)

    showReports(deltat)

    print("Press Enter to after some time has passed to read the temperatures again.")
    print("The temperatures should slowly be rising.")
    input()
    showReports(deltat)

    print("Press Enter to turn both heaters off")
    input()
    deltat.heaterOff(0)
    deltat.heaterOff(1)
    showReports(deltat)



def showReports(deltat):
    report = deltat.getReport(0)
    print("Primary Heater Report:")
    print(report)
    print()

    report = deltat.getReport(1)
    print("Secondary Heater Report:")
    print(report)
    print()

    for i in range(5):
        tempC = deltat.getTemperature(i)
        if tempC is None:
            print("Temp sensor %d: not connected" % i)
        else:
            print("Temp sensor %d: %.2f C" % (i, tempC))
    print()


if __name__ == "__main__":
    main()
